<?php
declare(strict_types=1);
namespace ParagonIE\PAST\Protocol;

use ParagonIE\ConstantTime\{
    Base64UrlSafe,
    Binary
};
use ParagonIE\PAST\Keys\{
    AsymmetricPublicKey,
    AsymmetricSecretKey,
    SymmetricKey
};
use ParagonIE\PAST\{
    ProtocolInterface,
    Util
};
use phpseclib\Crypt\RSA;

/**
 * Class Version1
 * @package ParagonIE\PAST\Protocol
 */
class Version1 implements ProtocolInterface
{
    const HEADER = 'v1';
    const CIPHER_MODE = 'aes-256-ctr';
    const HASH_ALGO = 'sha384';

    const NONCE_SIZE = 32;
    const MAC_SIZE = 48;
    const SIGN_SIZE = 256; // 2048-bit RSA = 256 byte signature

    /** @var RSA */
    protected static $rsa;

    /**
     * Encrypt a message using a shared key.
     *
     * @param string $data
     * @param SymmetricKey $key
     * @param string $footer
     * @param string $nonceForUnitTesting
     * @return string
     * @throws \Error
     * @throws \TypeError
     */
    public static function encrypt(
        string $data,
        SymmetricKey $key,
        string $footer = '',
        string $nonceForUnitTesting = ''
    ): string {
        return self::aeadEncrypt(
            $data,
            self::HEADER . '.local.',
            $key,
            $footer,
            $nonceForUnitTesting
        );
    }

    /**
     * Decrypt a message using a shared key.
     *
     * @param string $data
     * @param SymmetricKey $key
     * @param string $footer
     * @return string
     * @throws \Exception
     * @throws \Error
     * @throws \Exception
     * @throws \TypeError
     */
    public static function decrypt(string $data, SymmetricKey $key, string $footer = ''): string
    {
        return self::aeadDecrypt(
            Util::validateAndRemoveFooter($data, $footer),
            self::HEADER . '.local.',
            $key,
            $footer
        );
    }

    /**
     * Sign a message. Public-key digital signatures.
     *
     * @param string $data
     * @param AsymmetricSecretKey $key
     * @param string $footer
     * @return string
     */
    public static function sign(string $data, AsymmetricSecretKey $key, string $footer = ''): string
    {
        $header = self::HEADER . '.public.';
        $rsa = self::getRsa();
        $rsa->loadKey($key->raw());
        $signature = $rsa->sign(
            Util::preAuthEncode([$header, $data, $footer])
        );
        if ($footer) {
            return $header .
                Base64UrlSafe::encodeUnpadded($data . $signature) .
                '.' .
                Base64UrlSafe::encodeUnpadded($footer);
        }
        return $header . Base64UrlSafe::encodeUnpadded($data . $signature);
    }

    /**
     * Verify a signed message. Public-key digital signatures.
     *
     * @param string $signMsg
     * @param AsymmetricPublicKey $key
     * @param string $footer
     * @return string
     * @throws \Exception
     * @throws \TypeError
     */
    public static function verify(string $signMsg, AsymmetricPublicKey $key, string $footer = ''): string
    {
        $signMsg = Util::validateAndRemoveFooter($signMsg, $footer);
        $expectHeader = self::HEADER . '.public.';
        $givenHeader = Binary::safeSubstr($signMsg, 0, 10);
        if (!\hash_equals($expectHeader, $givenHeader)) {
            throw new \Exception('Invalid message header.');
        }
        $decoded = Base64UrlSafe::decode(Binary::safeSubstr($signMsg, 10));
        $len = Binary::safeStrlen($decoded);
        $message = Binary::safeSubstr($decoded, 0, $len - self::SIGN_SIZE);
        $signature = Binary::safeSubstr($decoded, $len - self::SIGN_SIZE);

        $rsa = self::getRsa();
        $rsa->loadKey($key->raw());
        $valid = $rsa->verify(
            Util::preAuthEncode([$givenHeader, $message, $footer]),
            $signature
        );
        if (!$valid) {
            throw new \Exception('Invalid signature for this message');
        }
        return $message;
    }

    /**
     * Authenticated Encryption with Associated Data -- Encryption
     *
     * Algorithm: AES-256-CTR + HMAC-SHA384 (Encrypt then MAC)
     *
     * @param string $plaintext
     * @param string $header
     * @param SymmetricKey $key
     * @param string $footer
     * @param string $nonceForUnitTesting
     * @return string
     * @throws \Error
     * @throws \TypeError
     */
    public static function aeadEncrypt(
        string $plaintext,
        string $header,
        SymmetricKey $key,
        string $footer = '',
        string $nonceForUnitTesting = ''
    ): string {
        if ($nonceForUnitTesting) {
            $nonce = self::getNonce($plaintext, $nonceForUnitTesting);
        } else {
            $nonce = self::getNonce($plaintext, \random_bytes(self::NONCE_SIZE));
        }
        list($encKey, $authKey) = $key->split(
            Binary::safeSubstr($nonce, 0, 16)
        );
        /** @var string $ciphertext */
        $ciphertext = \openssl_encrypt(
            $plaintext,
            self::CIPHER_MODE,
            $encKey,
            OPENSSL_RAW_DATA,
            Binary::safeSubstr($nonce, 16, 16)
        );
        if (!\is_string($ciphertext)) {
            throw new \Error('Encryption failed.');
        }
        $mac = \hash_hmac(
            self::HASH_ALGO,
            Util::preAuthEncode([$header, $nonce, $ciphertext, $footer]),
            $authKey,
            true
        );
        if ($footer) {
            return $header .
                Base64UrlSafe::encodeUnpadded($nonce . $ciphertext . $mac) .
                '.' .
                Base64UrlSafe::encodeUnpadded($footer);
        }
        return $header . Base64UrlSafe::encodeUnpadded($nonce . $ciphertext . $mac);
    }

    /**
     * Authenticated Encryption with Associated Data -- Decryption
     *
     * @param string $message
     * @param string $header
     * @param SymmetricKey $key
     * @param string $footer
     * @return string
     * @throws \Error
     * @throws \Exception
     * @throws \TypeError
     */
    public static function aeadDecrypt(
        string $message,
        string $header,
        SymmetricKey $key,
        string $footer = ''
    ): string {
        $expectedLen = Binary::safeStrlen($header);
        $givenHeader = Binary::safeSubstr($message, 0, $expectedLen);
        if (!\hash_equals($header, $givenHeader)) {
            throw new \Exception('Invalid message header.');
        }
        $decoded = Base64UrlSafe::decode(Binary::safeSubstr($message, $expectedLen));
        $len = Binary::safeStrlen($decoded);
        $nonce = Binary::safeSubstr($decoded, 0, self::NONCE_SIZE);
        $ciphertext = Binary::safeSubstr(
            $decoded,
            self::NONCE_SIZE,
            $len - (self::NONCE_SIZE + self::MAC_SIZE)
        );
        $mac = Binary::safeSubstr($decoded, $len - self::MAC_SIZE);

        list($encKey, $authKey) = $key->split(
            Binary::safeSubstr($nonce, 0, 16)
        );

        $calc = \hash_hmac(
            self::HASH_ALGO,
            Util::preAuthEncode([$header, $nonce, $ciphertext, $footer]),
            $authKey,
            true
        );
        if (!\hash_equals($calc, $mac)) {
            throw new \Exception('Invalid MAC');
        }

        /** @var string $plaintext */
        $plaintext = \openssl_decrypt(
            $ciphertext,
            self::CIPHER_MODE,
            $encKey,
            OPENSSL_RAW_DATA,
            Binary::safeSubstr($nonce, 16, 16)
        );
        if (!\is_string($plaintext)) {
            throw new \Error('Encryption failed.');
        }

        return $plaintext;
    }

    /**
     * Calculate a nonce from the message and a random nonce.
     * Mitigation against nonce-misuse.
     *
     * @param string $m
     * @param string $n
     * @return string
     * @throws \TypeError
     */
    public static function getNonce(string $m, string $n): string
    {
        $nonce = \hash_hmac(self::HASH_ALGO, $m, $n, true);
        return Binary::safeSubstr($nonce, 0, 32);
    }

    /**
     * Get the PHPSecLib RSA provider for signing
     *
     * @return RSA
     */
    public static function getRsa(): RSA
    {
        $rsa = new RSA();
        $rsa->setHash('sha384');
        $rsa->setMGFHash('sha384');
        $rsa->setEncryptionMode(RSA::SIGNATURE_PSS);
        return $rsa;
    }

    /**
     * @param string $keyData
     * @return string
     */
    public static function RsaGetPublicKey(string $keyData): string
    {
        $res = \openssl_pkey_get_private($keyData);
        /** @var array<string, string> $pubkey */
        $pubkey = \openssl_pkey_get_details($res);
        return \rtrim(
            \str_replace("\n", "\r\n", $pubkey['key']),
            "\r\n"
        );
    }
}
